@@ -18,6 +18,7 @@ from pygments.token import Keyword, Operator, Number, Name, Error, Comment, Toke
 from prompt_toolkit import CommandLine
 from prompt_toolkit.code import Completion, Code
 from prompt_toolkit.enums import LineMode
+from prompt_toolkit.history import FileHistory, History
 from prompt_toolkit.inputstream_handler import ViInputStreamHandler, EmacsInputStreamHandler, ViMode
 from prompt_toolkit.line import Exit, Line
 from prompt_toolkit.prompt import Prompt
@@ -127,7 +128,10 @@ class PythonViInputStreamHandler(_PythonInputStreamHandlerMixin, ViInputStreamHa
     def enter(self):
         self._auto_enable_multiline()
 
-        if self._line.multiline:
+        if self._line.mode == LineMode.INCREMENTAL_SEARCH:
+            self._line.exit_isearch(restore_original_line=False)
+
+        elif self._line.multiline:
             if self._vi_mode == ViMode.NAVIGATION:
                 # We are in VI-navigation mode after pressing `Meta`.
                 self._line.return_input()
@@ -142,8 +146,12 @@ class PythonEmacsInputStreamHandler(_PythonInputStreamHandlerMixin, EmacsInputSt
     def enter(self):
         self._auto_enable_multiline()
 
-        if self._line.multiline:
+        if self._line.mode == LineMode.INCREMENTAL_SEARCH:
+            self._line.exit_isearch(restore_original_line=False)
+
+        elif self._line.multiline:
             self._line.newline()
+
         else:
             # In single line input, always execute when pressing enter.
             self._line.return_input()
@@ -300,14 +308,17 @@ class PythonPrompt(Prompt):
     def _get_signature_tokens(self):
         result = []
         append = result.append
-        script = self.code._get_jedi_script()
         Signature = Token.Signature
+        script = self.code._get_jedi_script()
 
         # Show signatures in help text.
-        try:
-            signatures = script.call_signatures()
-        except ValueError:
-            # e.g. in case of an invalid \x escape.
+        if script:
+            try:
+                signatures = script.call_signatures()
+            except ValueError:
+                # e.g. in case of an invalid \x escape.
+                signatures = []
+        else:
             signatures = []
 
         if signatures:
@@ -337,17 +348,26 @@ class PythonPrompt(Prompt):
         append((Token, '\n  '))
         append((TB, '  '))
 
-        if self._pythonline.vi_mode:
+        # Mode
+        if self.line.mode == LineMode.INCREMENTAL_SEARCH:
+            append((TB.Mode, '(SEARCH)'))
+            append((TB, '  '))
+        elif self._pythonline.vi_mode:
             mode = self._pythonline._inputstream_handler._vi_mode
             if mode == ViMode.NAVIGATION:
-                append((TB.Mode, '(NAV)    '))
-                append((TB, '    '))
+                append((TB.Mode, '(NAV)'))
+                append((TB, '     '))
             elif mode == ViMode.INSERT:
-                append((TB.Mode, '(INSERT) '))
-                append((TB, ' '))
+                append((TB.Mode, '(INSERT)'))
+                append((TB, '  '))
             elif mode == ViMode.REPLACE:
                 append((TB.Mode, '(REPLACE)'))
                 append((TB, ' '))
+
+            if self._pythonline._inputstream_handler.is_recording_macro:
+                append((TB.Mode, 'recording'))
+                append((TB, ' '))
+
         else:
             append((TB.Mode, '(emacs)'))
             append((TB, ' '))
@@ -355,20 +375,25 @@ class PythonPrompt(Prompt):
         # Position in history.
         result.append((TB, '%i/%i ' % (self.line._working_index + 1, len(self.line._working_lines))))
 
-        if self.line.paste_mode:
-            append((TB.On, '[F6] Paste mode. (on)  '))
+        # Shortcuts.
+        if self.line.mode == LineMode.INCREMENTAL_SEARCH:
+            append((TB, '[Ctrl-G] Cancel search'))
         else:
-            append((TB.Off, '[F6] Paste mode. (off) '))
+            if self.line.paste_mode:
+                append((TB.On, '[F6] Paste mode (on)  '))
+            else:
+                append((TB.Off, '[F6] Paste mode (off) '))
 
-        if self.line.multiline:
-            append((TB.On, '[F7] Multiline (on)  '))
-        else:
-            append((TB.Off, '[F7] Multiline (off) '))
+            if self.line.multiline:
+                append((TB.On, '[F7] Multiline (on)  '))
+            else:
+                append((TB.Off, '[F7] Multiline (off) '))
+
+            if self.line.multiline:
+                append((TB, '[Meta+Enter] Execute'))
+            else:
+                append((TB, '                    '))
 
-        if self.line.multiline:
-            append((TB, '[Meta+Enter] Execute.'))
-        else:
-            append((TB, '                    '))
         append((TB, '  '))
 
         return result
@@ -413,24 +438,40 @@ class PythonCode(Code):
         return result
 
     def _get_jedi_script(self):
-        return jedi.Interpreter(self.text,
-                column=self.document.cursor_position_col,
-                line=self.document.cursor_position_row + 1,
-                path='input-text',
-                namespaces=[ self._locals, self._globals ])
+        try:
+            return jedi.Interpreter(self.text,
+                    column=self.document.cursor_position_col,
+                    line=self.document.cursor_position_row + 1,
+                    path='input-text',
+                    namespaces=[ self._locals, self._globals ])
+
+        except jedi.common.MultiLevelStopIteration:
+            # This happens when the document is just a backslash.
+            return None
+        except ValueError:
+            # Invalid cursor position.
+            # ValueError('`column` parameter is not in a valid range.')
+            return None
 
     def get_completions(self, recursive=False):
         """ Ask jedi to complete. """
         script = self._get_jedi_script()
 
-        for c in script.completions():
-            yield Completion(c.name, c.complete)
+        if script:
+            for c in script.completions():
+                yield Completion(c.name, c.complete)
 
 
 class PythonCommandLine(CommandLine):
     line_cls = PythonLine
     style_cls = PythonStyle
 
+    def history_cls(self):
+        if self.history_filename:
+            return FileHistory(self.history_filename)
+        else:
+            return History()
+
     @property
     def inputstream_handler_cls(self):
         if self.vi_mode:
@@ -438,10 +479,11 @@ class PythonCommandLine(CommandLine):
         else:
             return PythonEmacsInputStreamHandler
 
-    def __init__(self, globals=None, locals=None, vi_mode=False, stdin=None, stdout=None):
+    def __init__(self, globals=None, locals=None, vi_mode=False, stdin=None, stdout=None, history_filename=None):
         self.globals = globals or {}
         self.globals.update({ k: getattr(__builtins__, k) for k in dir(__builtins__) })
         self.locals = locals or {}
+        self.history_filename = history_filename
 
         self.vi_mode = vi_mode
 
@@ -510,7 +552,7 @@ class PythonCommandLine(CommandLine):
         print('\rKeyboardInterrupt')
 
 
-def embed(globals=None, locals=None, vi_mode=False):
+def embed(globals=None, locals=None, vi_mode=False, history_filename=None):
     """
     Call this to embed  Python shell at the current point in your program.
     It's similar to `IPython.embed` and `bpython.embed`. ::
@@ -520,4 +562,5 @@ def embed(globals=None, locals=None, vi_mode=False):
 
     :param vi_mode: Boolean. Use Vi instead of Emacs key bindings.
     """
-    PythonCommandLine(globals, locals, vi_mode=vi_mode).start_repl()
+    cli = PythonCommandLine(globals, locals, vi_mode=vi_mode, history_filename=history_filename)
+    cli.start_repl()
