@@ -13,12 +13,61 @@ __all__ = (
 )
 
 
-class Dynamic(object):
+class Dynamic(object): # TODO: inherit from Prompt itself.
     """
     Dynamic part in a prompt.
+
+    (This is used for parts of the prompt that are either changing all the
+    time, or for parts that are too heavy to build in the main thread.)
     """
-    def callback(self):
-        yield (Token.Prompt, '<dynamic>')
+    def __init__(self):
+        self.stopped = False
+
+    def in_thread(self, update_callback):
+        #wait(200ms)
+
+        if self.stopped:
+            return
+
+        else:
+            do_heavy_stuff()
+            update_callback()
+
+    def get_tokens(self):
+        """
+        Tokens to return while we're busy.
+        """
+        return [ (Token.Prompt, '<dynamic>') ]
+
+        #return [ Loader() ]
+
+
+class Loader(Dynamic):
+    def __init__(self):
+        self._state = 0
+
+    def in_thread(self, update_callback):
+        while True:
+            time.sleep(.3)
+            self._state = (self._state + 1) % 3
+            update_callback()
+
+    def get_tokens(self):
+        text = ['.  ', '.. ', '...'][self._state]
+        return [ (Token, 'Loading %s' % text) ]
+
+
+class Counter(Dynamic):
+    def in_thread(self, update_callback):
+        self.text = 0
+
+        while not self.stopped:
+            self.text += 1
+            update_callback()
+            time.sleep(.3)
+
+    def get_tokens(self):
+        return [ (Token, self.text) ]
 
 
 class PromptBase(object):
@@ -99,7 +148,7 @@ class Prompt(PromptBase):
         """
         yield (Token.Prompt.ISearch.Bracket, '(')
 
-        if self.line.isearch_direction == IncrementalSearchDirection.BACKWARD:
+        if self.line.isearch_state.isearch_direction == IncrementalSearchDirection.BACKWARD:
             yield (Token.Prompt.ISearch, 'reverse-i-search')
         else:
             yield (Token.Prompt.ISearch, 'i-search')
